import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

from datayoga_core import expression, utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.result import BlockResult

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.properties = utils.format_block_properties(self.properties)

        self.fields = {}
        for prop in self.properties["fields"]:
            self.fields[prop["field"]] = expression.compile(
                prop["language"],
                prop["expression"])

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")
        for field in self.fields:
            expression_results = self.fields[field].search_bulk(data)
            field_path = utils.split_field(field)

            for i, row in enumerate(data):
                obj = row
                # handle nested fields. in that case, the obj points at the nested entity
                for key in field_path[:-1]:
                    key = utils.unescape_field(key)
                    if key in obj:
                        obj = obj[key]
                    else:
                        obj[key] = {}
                # assign the new values
                obj[utils.unescape_field(field_path[-1:][0])] = expression_results[i]

        return utils.all_success(data)



import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional

from datayoga_core import utils
from datayoga_core.context import Context
from datayoga_core.job import Job
from datayoga_core.result import JobResult

logger = logging.getLogger("dy")


def compile(
        job_settings: Dict[str, Any],
        whitelisted_blocks: Optional[List[str]] = None) -> Job:
    """
    Compiles a job in YAML

    Args:
        job_settings (Dict[str, Any]): Job settings
        whitelisted_blocks: (Optional[List[str]], optional): Whitelisted blocks. Defaults to None.

    Returns:
        Job: Compiled job
    """
    logger.debug("Compiling job")
    return Job.compile(job_settings, whitelisted_blocks)


def validate(job_settings: Dict[str, Any], whitelisted_blocks: Optional[List[str]] = None):
    """
    Validates a job in YAML

    Args:
        job_settings (Dict[str, Any]): Job settings
        whitelisted_blocks: (Optional[List[str]], optional): Whitelisted blocks. Defaults to None.

    Raises:
        ValueError: When the job is invalid
    """
    logger.debug("Validating job")
    try:
        Job.validate(
            source=job_settings,
            whitelisted_blocks=whitelisted_blocks
        )
    except Exception as e:
        raise ValueError(e)


def transform(job_settings: Dict[str, Any],
              data: List[Dict[str, Any]],
              context: Optional[Context] = None,
              whitelisted_blocks: Optional[List[str]] = None) -> JobResult:
    """
    Transforms data against a certain job

    Args:
        job_settings (Dict[str, Any]): Job settings
        data (List[Dict[str, Any]]): Data to transform
        context (Optional[Context]): Context. Defaults to None.
        whitelisted_blocks: (Optional[List[str]]): Whitelisted blocks. Defaults to None.

    Returns:
        JobResult: Job result
    """
    job = compile(job_settings, whitelisted_blocks)
    job.init(context)
    logger.debug("Transforming data")
    return job.transform(data)


def get_connections_json_schema() -> Dict[str, Any]:
    # get the folder of all connection-specific schemas
    connection_schemas_folder = utils.get_resource_path(os.path.join("schemas", "connections"))
    # we traverse the json schemas for connection types
    schema_paths = Path(connection_schemas_folder).rglob("**/*.schema.json")
    connection_types = []
    connection_schemas = []
    for schema_path in schema_paths:
        connection_type = schema_path.name.split(".")[0]
        connection_types.append(connection_type)

        schema = utils.read_json(f"{schema_path}")
        # append to the array of allOf for the full schema
        connection_schemas.append({
            "if": {
                "properties": {
                    "type": {
                        "description": "Connection type",
                        "type": "string",
                        "const": connection_type
                    },
                },
                "required": ["type"]
            },
            "then": schema
        })

    connections_general_schema = utils.read_json(
        os.path.join(
            utils.get_bundled_dir() if utils.is_bundled() else os.path.dirname(os.path.realpath(__file__)),
            "resources", "schemas", "connections.schema.json"))
    connections_general_schema["patternProperties"]["."]["allOf"] = connection_schemas
    connections_general_schema["patternProperties"]["."]["properties"]["type"]["enum"] = connection_types

    return connections_general_schema



import copy
import os
import re
import sys
import uuid
from os import path
from typing import Any, Dict, List

import orjson
import yaml
from datayoga_core import result
from datayoga_core.block import Block
from datayoga_core.context import Context
from datayoga_core.expression import JMESPathExpression
from datayoga_core.result import BlockResult, Result, Status


def read_json(filename: str) -> Any:
    """
    Loads a filename as a JSON object

    Args:
        filename (str): JSON filename to load

    Returns:
        Any: JSON object
    """
    with open(filename, "r", encoding="utf8") as json_file:
        return orjson.loads(json_file.read())


def read_yaml(filename: str) -> Dict[str, Any]:
    """
    Loads a filename as a YAML object

    Args:
        filename (str): YAML filename to load

    Raises:
        ValueError: In case of invalid YAML

    Returns:
         Dict[str, Any]: YAML python object
    """
    try:
        with open(filename, "r", encoding="utf8") as yaml_file:
            return yaml.safe_load(yaml_file)
    except Exception as e:
        raise ValueError(f"Malformed YAML: {e}")


def format_block_properties(properties: Dict[str, Any]) -> Dict[str, Any]:
    """Adds `fields` array with the passed properties in case it's missing

    Args:
        properties (Dict[str, Any]): properties

    Returns:
        Dict[str, Any]: formatted properties with `fields` array
    """
    return {"fields": [properties]} if not "fields" in properties else properties


def is_bundled() -> bool:
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def get_bundled_dir() -> str:
    datayoga_dir = path.join(sys._MEIPASS, "datayoga_core")
    return datayoga_dir if os.path.isdir(datayoga_dir) else sys._MEIPASS


def get_resource_path(relative_path: str) -> str:
    if is_bundled():
        # we are running in a bundle
        return path.join(get_bundled_dir(), "resources", relative_path)
    else:
        # we are running in a normal Python environment
        return path.join(os.path.dirname(__file__), "resources", relative_path)


def split_field(field: str) -> List[str]:
    return re.split(r"(?<!\\)\.", field)


def unescape_field(field: str) -> str:
    return field.replace("\\.", ".")


def get_connection_details(connection_name: str, context: Context) -> Dict[str, Any]:
    if context and context.properties:
        connection = context.properties.get("connections", {}).get(connection_name)
        if connection:
            return connection

    raise ValueError(f"{connection_name} connection not found")


def all_success(records: List[Dict[str, Any]]) -> BlockResult:
    return BlockResult(processed=[Result(Status.SUCCESS, payload=row) for row in records])


def is_rejected(record: Dict[str, Any]) -> bool:
    return record.get(Block.RESULT_FIELD, result.SUCCESS).status == Status.REJECTED


def add_uid(record: Dict[str, Any]) -> Dict[str, Any]:
    return {Block.MSG_ID_FIELD: f"{uuid.uuid4()}", **record}


def remove_msg_id(record: dict) -> dict:
    return {k: v for k, v in record.items() if k != Block.MSG_ID_FIELD}


def explode_records(records: List[Dict[str, Any]], field_expression: str) -> List[Dict[str, Any]]:
    """
    Takes a list of records and a JMESPath expression specifying a field to be exploded, and returns a new list of records
    where each record has been "exploded" into multiple records based on the values in the specified field.

    Args:
        records (List[Dict[str, Any]]): A list of dictionaries representing records.
        field_expression (str): A string specifying the field to be exploded, in the form "field_name: expression".

    Returns:
        List[Dict[str, Any]]: A new list of dictionaries representing the exploded records.
    """
    field_name, expression = map(str.strip, field_expression.split(":", maxsplit=1))

    jmespath_expr = JMESPathExpression()
    jmespath_expr.compile(expression)

    exploded_records = []

    for record in records:
        # Apply the JMESPath expression to the current record to obtain the values to be exploded.
        field_values = jmespath_expr.search(record)

        # If the JMESPath expression returned any values, create a new record for each value and add it to the output list.
        if field_values:
            for field_value in field_values:
                new_record = copy.deepcopy(record)
                new_record[field_name] = field_value
                exploded_records.append(new_record)

    return exploded_records



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional, Union

import orjson
from datayoga_core import expression
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.expression import Language
from datayoga_core.result import BlockResult, Result, Status

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    @staticmethod
    def _prepare_expression(language: Language, expr: Union[dict, str]) -> str:
        """Prepares the expression to compile.
        The sql expression will be dumped as json.
        For the jmespath the expression will be generated."""

        if language == Language.SQL:
            return orjson.dumps(expr).decode() if isinstance(expr, dict) else expr.strip()

        def prepare_key(key: str) -> str:
            return f'"{key}"' if " " in key else key

        # If there is an object here, we generate an expression.
        if isinstance(expr, dict):
            expr = ", ".join(f"{prepare_key(k)}: {v}" for k, v in expr.items())
            expr = f"{{{expr}}}"

        return expr.strip()

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        language = self.properties["language"]
        expression_prop = self._prepare_expression(language, self.properties["expression"])

        # jmespath expression for map block must be enclosed in { }
        if not (expression_prop.startswith("{") and expression_prop.endswith("}")):
            raise ValueError("map expression must be in a json-like format enclosed in { }")

        self.expression = expression.compile(language, expression_prop)

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")
        mapped_rows = self.expression.search_bulk(data)
        # we always add the internal fields back, in case they were dropped in the map operation
        for mapped_row, original_row in zip(mapped_rows, data):
            original_msg_id = original_row.get(Block.MSG_ID_FIELD)
            original_opcode = original_row.get(Block.OPCODE_FIELD)
            if original_msg_id is not None:
                mapped_row[Block.MSG_ID_FIELD] = original_msg_id
            if original_opcode is not None:
                mapped_row[Block.OPCODE_FIELD] = original_opcode

        return BlockResult(processed=[Result(Status.SUCCESS, payload=row) for row in mapped_rows])



import logging
from enum import Enum, unique
from typing import Tuple

import sqlalchemy as sa
from datayoga_core import utils
from datayoga_core.context import Context

logger = logging.getLogger("dy")


@unique
class DbType(str, Enum):
    MYSQL = "mysql"
    ORACLE = "oracle"
    PSQL = "postgresql"
    SQLSERVER = "sqlserver"


DEFAULT_DRIVERS = {
    DbType.MYSQL: "mysql+pymysql",
    DbType.ORACLE: "oracle+oracledb",
    DbType.PSQL: "postgresql",
    DbType.SQLSERVER: "mssql+pymssql"
}


def get_engine(connection_name: str, context: Context, autocommit: bool = True) -> Tuple[sa.engine.Engine, DbType]:
    connection = utils.get_connection_details(connection_name, context)

    db_type = DbType(connection.get("type", "").lower())

    extra = {}

    if autocommit:
        extra["isolation_level"] = "AUTOCOMMIT"

    if db_type == DbType.ORACLE and connection.get("oracle_thick_mode", False):
        lib_dir = connection.get("oracle_thick_mode_lib_dir")
        extra["thick_mode"] = {"lib_dir": lib_dir} if lib_dir else {}

    engine = sa.create_engine(
        sa.engine.URL.create(
            drivername=connection.get("driver", DEFAULT_DRIVERS.get(db_type)),
            host=connection.get("host"),
            port=connection.get("port"),
            username=connection.get("user"),
            password=connection.get("password"),
            database=connection.get("database")),
        echo=connection.get("debug", False), connect_args=connection.get("connect_args", {}),
        **extra)

    return engine, db_type
